#!/usr/bin/env python2
from __future__ import print_function

"""filedrop.py: http upload target"""
__version__   = "0.8"
__author__    = "ed <filedrop@ocv.me>"
__credits__   = ["stackoverflow.com"]
__license__   = "MIT"
__copyright__ = 2019

# this occasionally corrupts files on certain python3 versions,
# currently determining which ones are buggy
#
# to reproduce the issue, start this program in one terminal:
#   python3 filedrop.py
#
# generate a 1.3MB file which is easy to diff:
#   hex="0 1 2 3 4 5 6 7 8 9"; row="$(for a in $hex; do printf ' %s.' $a; for b in $hex; do printf '%s' $b; done; done)"; for a in $hex; do for b in $hex; do for c in $hex; do for d in $hex; do printf '%s%s%s%s:%s\n' $a $b $c $d "$row"; done; done; done; done > digits.txt
#
# then start flooding it with that file (the awk at
# the end filters out unique hashes as they first appear):
#   while true; do curl -F 'f=@digits.txt' http://127.0.0.1:1234/; done | awk 'v[$0]++==0'
#
# alternatively use this client for more feedback:
#   fn=digits.txt; sum=$(sha1sum $fn | cut -c-40); printf "\ngood sha1 hash = %s\nnow flooding" $sum; while true; do curl -sF "f=@$fn" http://127.0.0.1:1234/; done | perl -MTime::HiRes=time -ne '$|++; chomp; $n++; printf "."; next if /:'$sum'$/; printf "\n%s #%s @%s\n", $_, $n, time()'
# 
# example good+corrupted hash of another file, "gm.dls":
#   len:3440660 sha1:57145f5f449624a10a357d64c982c6eaa7ebc1d9
#   len:3432615 sha1:ae0c9d979afdbca881bac8cc049372a6c491ff7a

import os
import sys
import time
import hashlib
import select

try:
	from BaseHTTPServer import BaseHTTPRequestHandler, HTTPServer
except:
	from http.server import BaseHTTPRequestHandler, HTTPServer


class Srv(BaseHTTPRequestHandler):
	def readbytes(self, f, num_bytes=4096):
		"""blocks until any data, returns [] on disconnect"""
		while True:
			r,w,e = select.select([f], [], [], 0)
			if f in r:
				return os.read(f.fileno(), num_bytes)
			
			time.sleep(0.01)
	
	def readline(self, f):
		"""reads one ascii \r\n line aggressively"""
		ret = b''
		while True:
			ret += f.read(2)
			if ret[-1:] == b'\r':
				ret += f.read(1)
			if ret[-2:] == b'\r\n':
				return ret[:-2]
	
	def reply(self, body):
		header = 'HTTP/1.1 200 OK\r\nContent-Type: text/html\r\nContent-Length: {0}\r\n\r\n'.format(len(body)).encode('ascii')
		self.wfile.write(header + body)
		return body
	
	def loud_reply(self, body):
		print(body.decode('ascii'), end='')
		self.reply(body)
	
	def do_GET(self):
		self.reply(b'<form method="post" enctype="multipart/form-data"><input type="file" name="f"><input type="submit">')
	
	def extract_filedata(self, boundary, ifd):
		blen = len(boundary)
		while True:
			buf = self.readbytes(ifd)
			if not buf:
				# abort: client disconnected
				print("CLIENT DISCONNECT 1")
				return
			
			while True:
				d = len(buf) - blen
				if d > 0:
					# buffer growing large; yield everything except
					# the part at the end (maybe start of boundary)
					yield buf[:d]
					buf = buf[d:]
				
				# look for boundary near the end of the buffer
				for n in range(1, len(buf)+1):
					if not buf[-n:] in boundary:
						n -= 1
						break
				
				if n == 0 \
				or not boundary.startswith(buf[-n:]):
					# no boundary contents near the buffer edge
					break
				
				if blen == n:
					# EOF: found boundary
					yield buf[:-n]
					return
				
				buf2 = self.readbytes(ifd)
				if not buf2:
					# abort: client disconnected
					print("CLIENT DISCONNECT 2")
					return
				
				buf += buf2
				
			yield buf
	
	def handle_multipart(self, boundary, ifd, ofd):
		tlen = 0
		hashobj = hashlib.sha1()
		for buf in self.extract_filedata(boundary, ifd):
			tlen += len(buf)
			hashobj.update(buf)
			ofd.write(buf)
		
		return tlen, hashobj.hexdigest()
	
	def do_POST(self):
		#for k, v in self.headers.items():
		#	print('[{0}] [{1}]'.format(k, v))

		try:
			if self.headers['expect'].lower() == '100-continue':
				self.wfile.write(b'HTTP/1.1 100 Continue\r\n\r\n')
		except:
			pass
		
		ln = self.readline(self.rfile)
		#print(ln)
		boundary = b'\r\n' + ln + b'--\r\n'
		while ln:
			#print(ln)
			ln = self.readline(self.rfile)
		
		fn = 'inc.{0:.3f}'.format(time.time())
		#print('writing to {0}'.format(fn))
		#fn = '/dev/null'
		with open(fn, 'wb') as f:
			sz, sha1 = self.handle_multipart(boundary, self.rfile, f)
		
		msg = 'len:{0} sha1:{1}\n'.format(sz, sha1)
		self.loud_reply(msg.encode('ascii'))


port = 1234
if len(sys.argv) > 1:
	port = int(sys.argv[1])

print('listening @ http://127.0.0.1:{0}/'.format(port))
HTTPServer(('0.0.0.0', port), Srv).serve_forever()
